--
-- $RCSfile: controller.vhd,v $
-- $Revision: 1.4 $
-- $Author: petera $
-- $Date: 90/05/29 14:19:12 $
--

use work.dp32_types.all;

entity controller is
  generic (Tpd : Time := unit_delay);
  port (phi1, phi2 : in bit;
      	reset : in bit;
      	opcode : in bit_8;
      	read, write, fetch : out bit;
        ready : in bit;
      	addr_latch_en : out bit;
	disp_latch_en : out bit;
	disp_out_en : out bit;
      	d2_en : out bit;
      	dr_en : out bit;
      	instr_latch_en : out bit;
      	immed_signext_en : out bit;
      	ALU_op : out ALU_command;
      	CC_latch_en : out bit;
      	CC_comp_result : in bit;
	PC_latch_en : out bit;
	PC_out_en : out bit;
      	reg_port1_en : out bit;
      	reg_port2_en : out bit;
      	reg_port3_en : out bit;
      	reg_port2_mux_sel : out bit;
	reg_res_latch_en : out bit);
end controller;

architecture behaviour of controller is
  
begin

  control_machine: process
  
    type controller_state is 
      (	resetting,
      	fetch_0, fetch_1, fetch_2,
      	decode,
	disp_fetch_0, disp_fetch_1, disp_fetch_2,
	execute_0, execute_1, execute_2);
    
    variable state, next_state : controller_state;
    variable write_back_pending : boolean;

    type ALU_op_select_table is array (natural range 0 to 255) of ALU_command;
    
    constant ALU_op_select : ALU_op_select_table :=
      ( 16#00# => add,
	16#01# => subtract,
	16#02# => multiply,
	16#03# => divide,
	16#10# => add,
	16#11# => subtract,
	16#12# => multiply,
	16#13# => divide,
	16#04# => log_and,
	16#05# => log_or,
	16#06# => log_xor,
	16#07# => log_mask,
	others => disable );
    
  begin
    --
    -- start of clock cycle
    --
    wait until phi1 = '1';
    --
    -- check for reset
    --
    if reset = '1' then
      state := resetting;
      --
      -- reset external bus signals
      --
      read <= '0' after Tpd;
      fetch <= '0' after Tpd;
      write <= '0' after Tpd;
      --
      -- reset dp32 internal control signals
      --
      addr_latch_en <= '0' after Tpd;
      disp_latch_en <= '0' after Tpd;
      disp_out_en <= '0' after Tpd;
      d2_en <= '0' after Tpd;
      dr_en <= '0' after Tpd;
      instr_latch_en <= '0' after Tpd;
      immed_signext_en <= '0' after Tpd;
      ALU_op <= disable after Tpd;
      CC_latch_en <= '0' after Tpd;
      PC_latch_en <= '0' after Tpd;
      PC_out_en <= '0' after Tpd;
      reg_port1_en <= '0' after Tpd;
      reg_port2_en <= '0' after Tpd;
      reg_port3_en <= '0' after Tpd;
      reg_port2_mux_sel <= '0' after Tpd;
      reg_res_latch_en <= '0' after Tpd;
      --
      -- clear write-back flag
      --
      write_back_pending := false;
      --
    else -- reset = '0'
      state := next_state;
    end if;
    --
    -- dispatch action for current state
    --
    case state is
      when resetting =>
     	--
      	-- check for reset going inactive at end of clock cycle
	--
      	wait until phi2 = '0';
	if reset = '0' then
	  next_state := fetch_0;
	else
	  next_state := resetting;
	end if;
	--
      when fetch_0 =>
      	--
	-- clean up after previous execute cycles
	--
	reg_port1_en <= '0' after Tpd;
	reg_port2_mux_sel <= '0' after Tpd;
	reg_port2_en <= '0' after Tpd;
	immed_signext_en <= '0' after Tpd;
	disp_out_en <= '0' after Tpd;
	dr_en <= '0' after Tpd;
	read <= '0' after Tpd;
	d2_en <= '0' after Tpd;
	write <= '0' after Tpd;
      	--
	-- handle pending register write-back
	--
	if write_back_pending then
	  reg_port3_en <= '1' after Tpd;
	end if;
	--
	-- enable PC via ALU to address latch
	--
	PC_out_en <= '1' after Tpd;	  -- enable PC onto op1_bus
	ALU_op <= pass1 after Tpd;	  -- pass PC to r_bus
	--
	wait until phi2 = '1';
	addr_latch_en <= '1' after Tpd;	  -- latch instr address
	wait until phi2 = '0';
	addr_latch_en <= '0' after Tpd;
	--
	next_state := fetch_1;
	--
      when fetch_1 =>
      	--
	-- clear pending register write-back
	--
	if write_back_pending then
	  reg_port3_en <= '0' after Tpd;
	  write_back_pending := false;
	end if;
	--
	-- increment PC & start bus read
	--
	ALU_op <= incr1 after Tpd;	  -- increment PC onto r_bus
	fetch <= '1' after Tpd;
	read <= '1' after Tpd;
	--
	wait until phi2 = '1';
	PC_latch_en <= '1' after Tpd;	  -- latch incremented PC
	wait until phi2 = '0';
	PC_latch_en <= '0' after Tpd;
	--
	next_state := fetch_2;
	--
      when fetch_2 =>
      	--
	-- cleanup after previous fetch_1
	--
	PC_out_en <= '0' after Tpd;	  -- disable PC from op1_bus
	ALU_op <= disable after Tpd;	  -- disable ALU from r_bus
      	--
	-- latch current instruction
	--
	dr_en <= '1' after Tpd;	  -- enable fetched instr onto r_bus
	--
	wait until phi2 = '1';
	instr_latch_en <= '1' after Tpd;  -- latch fetched instr from r_bus
	wait until phi2 = '0';
	instr_latch_en <= '0' after Tpd;
	--
	if ready = '1' then
	  next_state := decode;
	else
	  next_state := fetch_2;	  -- extend bus read
	end if;
      when decode =>
      	--
	-- terminate bus read from previous fetch_2
	--
	fetch <= '0' after Tpd;
	read <= '0' after Tpd;
	dr_en <= '0' after Tpd;		  -- disable fetched instr from r_bus
	--
	-- delay to allow decode logic to settle
	--
	wait until phi2 = '0';
	--
	-- next state based on opcode of currect instruction
	--
	case opcode is
	  when op_add | op_sub | op_mul | op_div 
	      | op_addq | op_subq | op_mulq | op_divq
	      | op_land | op_lor | op_lxor | op_lmask
	      | op_ldq | op_stq =>
	    next_state := execute_0;
	  when op_ld | op_st =>
	    next_state := disp_fetch_0;	  -- fetch offset
	  when op_br | op_bi =>
	    if CC_comp_result = '1' then  -- if branch taken
	      next_state := disp_fetch_0; --   fetch displacement
	    else			  -- else
	      next_state := execute_0;	  --   increment PC past displacement
	    end if;
	  when op_brq | op_biq =>
	    if CC_comp_result = '1' then  -- if branch taken
	      next_state := execute_0;	  --   add immed displacement to PC
	    else			  -- else
	      next_state := fetch_0;	  --   no action needed
	    end if;
	  when others =>
	    assert false report "illegal instruction" severity warning;
	    next_state := fetch_0;	  -- ignore and carry on
	end case; -- op
	--
      when disp_fetch_0 =>
      	--
	-- enable PC via ALU to address latch
	--
	PC_out_en <= '1' after Tpd;	  -- enable PC onto op1_bus
	ALU_op <= pass1 after Tpd;	  -- pass PC to r_bus
	--
	wait until phi2 = '1';
	addr_latch_en <= '1' after Tpd;	  -- latch displacement address
	wait until phi2 = '0';
	addr_latch_en <= '0' after Tpd;
	--
	next_state := disp_fetch_1;
      	--
      when disp_fetch_1 =>
      	--
	-- increment PC & start bus read
	--
	ALU_op <= incr1 after Tpd;	  -- increment PC onto r_bus
	fetch <= '1' after Tpd;
	read <= '1' after Tpd;
	--
	wait until phi2 = '1';
	PC_latch_en <= '1' after Tpd;	  -- latch incremented PC
	wait until phi2 = '0';
	PC_latch_en <= '0' after Tpd;
	--
	next_state := disp_fetch_2;
	--
      when disp_fetch_2 =>
      	--
	-- cleanup after previous disp_fetch_1
	--
	PC_out_en <= '0' after Tpd;	  -- disable PC from op1_bus
	ALU_op <= disable after Tpd;	  -- disable ALU from r_bus
      	--
	-- latch displacement
	--
	wait until phi2 = '1';
	disp_latch_en <= '1' after Tpd;  -- latch fetched disp from r_bus
	wait until phi2 = '0';
	disp_latch_en <= '0' after Tpd;
	--
	if ready = '1' then
	  next_state := execute_0;
	else
	  next_state := disp_fetch_2;	  -- extend bus read
	end if;
      when execute_0 =>
      	--
	-- terminate bus read from previous disp_fetch_2
	--
	fetch <= '0' after Tpd;
	read <= '0' after Tpd;
	--
      	case opcode is
	  when op_add | op_sub | op_mul | op_div
	      | op_addq | op_subq | op_mulq | op_divq 
	      | op_land | op_lor | op_lxor | op_lmask =>
	    --
	    reg_port1_en <= '1' after Tpd; -- enable r1 onto op1_bus
	    if opcode = op_addq or opcode = op_subq or 
	      	opcode = op_mulq or opcode = op_divq then
	      immed_signext_en <= '1' after Tpd; -- enable i8 onto op2_bus
	    else
	      reg_port2_mux_sel <= '0' after Tpd; -- select a2 as port2 address
	      reg_port2_en <= '1' after Tpd; -- enable r2 onto op2_bus
	    end if;
	    -- select ALU operation
	    ALU_op <= ALU_op_select(bit_8_to_int(opcode)) after Tpd;
	    --
	    wait until phi2 = '1';
	    CC_latch_en <= '1' after Tpd; -- latch cond codes from ALU
	    reg_res_latch_en <= '1' after Tpd; -- latch result for reg write
	    wait until phi2 = '0';
	    CC_latch_en <= '0' after Tpd;
	    reg_res_latch_en <= '0' after Tpd;
	    --
	    next_state := fetch_0;	  -- execution complete
	    write_back_pending := true;	  -- register write_back required
      	    --
	  when op_ld | op_st | op_ldq | op_stq =>
	    --
	    reg_port1_en <= '1' after Tpd; -- enable r1 to op1_bus
	    if opcode = op_ld or opcode = op_st then
	      disp_out_en <= '1' after Tpd; -- enable displacement to op2_bus
	    else
	      immed_signext_en <= '1' after Tpd; -- enable i8 to op2_bus
	    end if;
	    ALU_op <= add after Tpd;	  -- effective address to r_bus
	    --
	    wait until phi2 = '1';
	    addr_latch_en <= '1' after Tpd; -- latch effective address
	    wait until phi2 = '0';
	    addr_latch_en <= '0' after Tpd;
	    --
	    next_state := execute_1;
	    --
	  when op_br | op_bi | op_brq | op_biq =>
	    --
	    if CC_comp_result = '1' then
	      if opcode = op_br then
	      	PC_out_en <= '1' after Tpd;
	      	disp_out_en <= '1' after Tpd;
	      elsif opcode = op_bi then
	      	reg_port1_en <= '1' after Tpd;
		disp_out_en <= '1' after Tpd;
	      elsif opcode = op_brq then
	      	PC_out_en <= '1' after Tpd;
		immed_signext_en <= '1' after Tpd;
	      else -- opcode = op_biq
	      	reg_port1_en <= '1' after Tpd;
		immed_signext_en <= '1' after Tpd;
	      end if;
	      ALU_op <= add after Tpd;
	    else
	      assert opcode = op_br or opcode = op_bi
	      	report "reached state execute_0 when brq or biq not taken"
		severity error;
	      PC_out_en <= '1' after Tpd;
	      ALU_op <= incr1 after Tpd;
	    end if;
	    --
	    wait until phi2 = '1';
	    PC_latch_en <= '1' after Tpd; -- latch incremented PC
	    wait until phi2 = '0';
	    PC_latch_en <= '0' after Tpd;
	    --
	    next_state := fetch_0;
	    --
	  when others =>
	    null;
	end case; -- op
      	--
      when execute_1 =>
      	--
	-- opcode is load or store instruction.
	-- cleanup after previous execute_0
	--
	reg_port1_en <= '0' after Tpd;
	if opcode = op_ld or opcode = op_st then
	  disp_out_en <= '0' after Tpd; -- disable displacement from op2_bus
	else
	  immed_signext_en <= '0' after Tpd; -- disable i8 from op2_bus
	end if;
	ALU_op <= add after Tpd;	  -- disable ALU from r_bus
	--
	-- start bus cycle
	--
	if opcode = op_ld or opcode = op_ldq then
	  fetch <= '0' after Tpd;	  -- start bus read
	  read <= '1' after Tpd;
	else -- opcode = op_st or opcode = op_stq
	  reg_port2_mux_sel <= '1' after Tpd; -- address a3 to port2
	  reg_port2_en <= '1' after Tpd;  -- reg port2 to op2_bus
	  d2_en <= '1' after Tpd;	  -- enable op2_bus to d_bus buffer
	  write <= '1' after Tpd;	  -- start bus write
	end if;
	--
	next_state := execute_2;
      	--
      when execute_2 =>
      	--
	-- opcode is load or store instruction.
	-- for load, enable read data onto r_bus
	--
	if opcode = op_ld or opcode = op_ldq then
	  dr_en <= '1' after Tpd;	  -- enable data to r_bus
	  wait until phi2 = '1';
	  reg_res_latch_en <= '1' after Tpd; -- latch data in reg result latch
	  wait until phi2 = '0';
	  reg_res_latch_en <= '0' after Tpd;
	  write_back_pending := true;	  -- write-back pending
	end if;
	--
	next_state := fetch_0;
      	--
    end case; -- state
  end process control_machine;

end behaviour;
